from setuptools import setup

setup(
    name='mtcnn-pytorch',
    version='1.0.1',
    packages=['mtcnn'],
    package_data={'mtcnn': ['weights/*.npy']},
    url='',
    license='',
    author='Dan Antoshchenko',
    author_email='antoshchenko.b@gmail.com',
    description=''
)
